@@ -1,7 +1,6 @@
 #ifndef BON_FILE
 #define BON_FILE
 
-#include <stdio.h>
 #include "bonnie.h"
 class Semaphore;
 class BonTimer;
@@ -9,8 +8,8 @@ class BonTimer;
 class CFileOp
 {
 public:
-  CFileOp(BonTimer &timer, int file_size, int chunk_bits, bool use_sync = false);
-  int open(CPCCHAR basename, bool create, bool use_fopen = false);
+  CFileOp(BonTimer &timer, int file_size, int chunk_bits, bool use_sync = false, bool use_direct_io = false);
+  int open(CPCCHAR base_name, bool create, bool use_fopen = false);
   ~CFileOp();
   int write_block_putc();
   int write_block(PVOID buf);
@@ -26,26 +25,24 @@ public:
   BonTimer &getTimer() { return m_timer; }
   int chunks() const { return m_total_chunks; }
 private:
-  int m_open(CPCCHAR basename, int ind, bool create);
+  int m_open(CPCCHAR base_name, int ind, bool create);
+
+  BonTimer &m_timer;
+  FILE **m_stream;
+  FILE_TYPE *m_fd;
+  bool m_isopen;
+  char *m_name;
+  bool m_sync;
+  bool m_use_direct_io;
   const int m_chunk_bits, m_chunk_size;
   int m_chunks_per_file, m_total_chunks;
   int m_last_file_chunks;
   int m_cur_pos;
   int m_file_ind;
-  BonTimer &m_timer;
-  FILE **m_stream;
-#ifdef OS2
-  HFILE *m_fd;
-#else
-  int *m_fd;
-#endif
-  int m_num_files;
   int m_file_size;
-  bool m_isopen;
-  char *m_name;
+  int m_num_files;
   CFileOp(const CFileOp &f);
   CFileOp & operator =(const CFileOp &f);
-  bool m_sync;
   char *m_buf;
 };
 
