@@ -42,6 +42,12 @@
 #include <sys/utsname.h>
 #include <signal.h>
 
+#ifdef AIX_MEM_SIZE
+#include <cf.h>
+#include <sys/cfgodm.h>
+#include <sys/cfgdb.h>
+#endif
+
 void usage();
 
 class CGlobalItems
@@ -50,6 +56,7 @@ public:
   bool quiet;
   bool fast;
   bool sync_bonnie;
+  bool use_direct_io;
   BonTimer timer;
   int ram;
   Semaphore sem;
@@ -59,9 +66,10 @@ public:
   int chunk_size() const { return m_chunk_size; }
   bool *doExit;
   void set_chunk_size(int size)
-    { delete m_buf; m_buf = new char[size]; m_chunk_size = size; }
+    { delete m_buf; pa_new(size, m_buf, m_buf_pa); m_chunk_size = size; }
 
-  char *buf() { return m_buf; }
+  // Return the page-aligned version of the local buffer
+  char *buf() { return m_buf_pa; }
 
   CGlobalItems(bool *exitFlag);
   ~CGlobalItems() { delete name; delete m_buf; }
@@ -83,27 +91,48 @@ public:
   }
 private:
   int m_chunk_size;
-  char *m_buf;
-
+  char *m_buf;     // Pointer to the entire buffer
+  char *m_buf_pa;  // Pointer to the page-aligned version of the same buffer
 
   CGlobalItems(const CGlobalItems &f);
   CGlobalItems & operator =(const CGlobalItems &f);
+
+  // Implement a page-aligned version of new.
+  // 'p' is the pointer created
+  // 'page_aligned_p' is the page-aligned pointer created
+  void pa_new(unsigned int num_bytes, char *&p, char *&page_aligned_p)
+  {
+#ifdef NON_UNIX
+    SYSTEM_INFO system_info;
+    GetSystemInfo(&system_info);
+    long page_size = system_info.dwPageSize;
+#else
+    int page_size = getpagesize();
+#endif
+    p = ::new char [num_bytes + page_size];
+
+    page_aligned_p = (char *)((((unsigned long)p + page_size - 1) / page_size) * page_size);
+  }
 };
 
 CGlobalItems::CGlobalItems(bool *exitFlag)
- : sem(SemKey, TestCount)
+ : quiet(false)
+ , fast(false)
+ , sync_bonnie(false)
+ , use_direct_io(false)
+ , timer()
+ , ram(0)
+ , sem(SemKey, TestCount)
+ , name(NULL)
+ , bufSync(false)
+ , chunk_bits(DefaultChunkBits)
+ , doExit(exitFlag)
+ , m_chunk_size(DefaultChunkSize)
+ , m_buf(NULL)
+ , m_buf_pa(NULL)
 {
-  quiet = false;
-  fast = false;
-  sync_bonnie = false;
-  ram = 0;
-  name = NULL;
+  pa_new(m_chunk_size, m_buf, m_buf_pa);
   SetName(".");
-  bufSync = false;
-  m_chunk_size = DefaultChunkSize;
-  m_buf = new char[m_chunk_size];
-  chunk_bits = DefaultChunkBits;
-  doExit = exitFlag;
 }
 
 void CGlobalItems::decrement_and_wait(int nr_sem)
@@ -117,14 +146,28 @@ int TestDirOps(int directory_size, int max_size, int min_size
 int TestFileOps(int file_size, CGlobalItems &globals);
 
 static bool exitNow;
+static bool already_printed_error;
 
-void ctrl_c_handler(int sig)
+#ifdef USE_SA_SIGACTION
+#define SIGNAL_NUMBER siginf->si_signo
+#else
+#define SIGNAL_NUMBER sig
+#endif
+
+extern "C"
 {
-  if(sig == SIGXCPU)
-    fprintf(stderr, "Exceeded CPU usage.\n");
-  else if(sig == SIGXFSZ)
-    fprintf(stderr, "exceeded file storage limits.\n");
-  exitNow = true;
+  void ctrl_c_handler(int sig
+#ifdef USE_SA_SIGACTION
+		    , siginfo_t *siginf, void *unused
+#endif
+		     )
+  {
+    if(SIGNAL_NUMBER == SIGXCPU)
+      fprintf(stderr, "Exceeded CPU usage.\n");
+    else if(SIGNAL_NUMBER == SIGXFSZ)
+      fprintf(stderr, "exceeded file storage limits.\n");
+    exitNow = true;
+  }
 }
 
 int main(int argc, char *argv[])
@@ -142,11 +185,16 @@ int main(int argc, char *argv[])
   bool setSize = false;
 
   exitNow = false;
+  already_printed_error = false;
 
   struct sigaction sa;
-  sa.sa_sigaction = NULL;
+#ifdef USE_SA_SIGACTION
+  sa.sa_sigaction = &ctrl_c_handler;
+  sa.sa_flags = SA_RESETHAND | SA_SIGINFO;
+#else
   sa.sa_handler = ctrl_c_handler;
   sa.sa_flags = SA_RESETHAND;
+#endif
   if(sigaction(SIGINT, &sa, NULL)
    || sigaction(SIGXCPU, &sa, NULL)
    || sigaction(SIGXFSZ, &sa, NULL))
@@ -154,6 +202,9 @@ int main(int argc, char *argv[])
     printf("Can't handle SIGINT.\n");
     return 1;
   }
+#ifdef USE_SA_SIGACTION
+  sa.sa_sigaction = NULL;
+#endif
   sa.sa_handler = SIG_IGN;
   if(sigaction(SIGHUP, &sa, NULL))
   {
@@ -168,10 +219,23 @@ int main(int argc, char *argv[])
   {
     globals.ram = page_size/1024 * (num_pages/1024);
   }
+#else
+
+#ifdef AIX_MEM_SIZE
+  struct CuAt *odm_obj;
+  int how_many;
+
+  odm_set_path("/etc/objrepos");
+  odm_obj = getattr("sys0", "realmem", 0, &how_many);
+  globals.ram = atoi(odm_obj->value) / 1024;
+  odm_terminate();
+  printf("Memory = %d MiB\n", globals.ram);
+#endif
+
 #endif
 
   int int_c;
-  while(-1 != (int_c = getopt(argc, argv, "bd:fg:m:n:p:qr:s:u:x:y")) )
+  while(-1 != (int_c = getopt(argc, argv, "bd:fg:m:n:p:qr:s:u:x:yD")) )
   {
     switch(char(int_c))
     {
@@ -183,7 +247,11 @@ int main(int argc, char *argv[])
         globals.bufSync = true;
       break;
       case 'd':
-        globals.SetName(optarg);
+        if(chdir(optarg))
+        {
+          fprintf(stderr, "Can't change to directory \"%s\".\n", optarg);
+          usage();
+        }
       break;
       case 'f':
         globals.fast = true;
@@ -238,7 +306,8 @@ int main(int argc, char *argv[])
           usage();
         userName = strdup(optarg);
         int i;
-        for(i = 0; userName[i] && userName[i] != ':'; i++);
+        for(i = 0; userName[i] && userName[i] != ':'; i++)
+        {}
         if(userName[i] == ':')
         {
           if(groupName)
@@ -256,6 +325,10 @@ int main(int argc, char *argv[])
                            defined semaphore */
         globals.sync_bonnie = true;
       break;
+      case 'D':
+                        /* open file descriptor with direct I/O */
+        globals.use_direct_io = true;
+      break;
     }
   }
   if(optind < argc)
@@ -279,7 +352,7 @@ int main(int argc, char *argv[])
 
   if(userName || groupName)
   {
-    if(bon_setugid(userName, groupName))
+    if(bon_setugid(userName, groupName, globals.quiet))
       return 1;
     if(userName)
       free(userName);
@@ -317,7 +390,8 @@ int main(int argc, char *argv[])
     usage();
   int i;
   globals.chunk_bits = 0;
-  for(i = globals.chunk_size(); i > 1; i = i >> 1, globals.chunk_bits++);
+  for(i = globals.chunk_size(); i > 1; i = i >> 1, globals.chunk_bits++)
+  {}
   if(1 << globals.chunk_bits != globals.chunk_size())
     usage();
 
@@ -325,7 +399,12 @@ int main(int argc, char *argv[])
      && (directory_max_size < directory_min_size || directory_max_size < 0
      || directory_min_size < 0) )
     usage();
-  if(file_size > IOFileSize * MaxIOFiles || file_size > (1 << (31 - 20 + globals.chunk_bits)) )
+  /* If the storage size is too big for the maximum number of files (1000G) */
+  if(file_size > IOFileSize * MaxIOFiles)
+    usage();
+  /* If the file size is so large and the chunk size is so small that we have
+   * more than 2G of chunks */
+  if(globals.chunk_bits < 20 && file_size > (1 << (31 - 20 + globals.chunk_bits)) )
     usage();
   // if doing more than one test run then we print a header before the
   // csv format output.
@@ -375,7 +454,7 @@ TestFileOps(int file_size, CGlobalItems &globals)
 {
   if(file_size)
   {
-    CFileOp file(globals.timer, file_size, globals.chunk_bits, globals.bufSync);
+    CFileOp file(globals.timer, file_size, globals.chunk_bits, globals.bufSync, globals.use_direct_io);
     int    num_chunks;
     int    words;
     char  *buf = globals.buf();
@@ -551,24 +630,24 @@ TestDirOps(int directory_size, int max_size, int min_size
   {
     return 0;
   }
-  // if directory_size (in K) * data per file*2 > ram << 10 (IE memory /1024)
+  // if directory_size (in K) * data per file*2 > (ram << 10) (IE memory /1024)
   // then the storage of file names will take more than half RAM and there
   // won't be enough RAM to have Bonnie++ paged in and to have a reasonable
   // meta-data cache.
-  if(globals.ram
-      && directory_size * MaxDataPerFile * 2 * num_directories
-                  > (globals.ram << 10))
+  if(globals.ram && directory_size * MaxDataPerFile * 2 > (globals.ram << 10))
   {
-    fprintf(stderr, "When testing %dK * %d of files in %d MB of RAM the system is likely to\n"
-            "start paging Bonnie++ data and the test will give suspect\n"
-            "results, use less files or install more RAM for this test.\n"
-           , directory_size, num_directories, globals.ram);
+    fprintf(stderr
+        , "When testing %dK of files in %d MiB of RAM the system is likely to\n"
+           "start paging Bonnie++ data and the test will give suspect\n"
+           "results, use less files or install more RAM for this test.\n"
+          , directory_size, globals.ram);
     return 1;
   }
-  if(directory_size * num_directories * MaxDataPerFile > (1 << 20))
+  // Can't use more than 1G of RAM
+  if(directory_size * MaxDataPerFile > (1 << 20))
   {
-    fprintf(stderr, "Not enough ram to test with %dK * %d files.\n"
-                  , directory_size, num_directories);
+    fprintf(stderr, "Not enough ram to test with %dK files.\n"
+                  , directory_size);
     return 1;
   }
   globals.decrement_and_wait(CreateSeq);
@@ -607,12 +686,12 @@ void
 usage()
 {
   fprintf(stderr,
-    "usage: Bonnie [-d scratch-dir] [-s size(Mb)[:chunk-size(b)]]\n"
-    "              [-n number-to-stat[:max-size[:min-size][:num-directories]]]\n"
-    "              [-m machine-name]\n"
-    "              [-r ram-size-in-Mb]\n"
-    "              [-x number-of-tests] [-u uid-to-use:gid-to-use] [-g gid-to-use]\n"
-    "              [-q] [-f] [-b] [-p processes | -y]\n"
+    "usage: bonnie++ [-d scratch-dir] [-s size(MiB)[:chunk-size(b)]]\n"
+    "                [-n number-to-stat[:max-size[:min-size][:num-directories]]]\n"
+    "                [-m machine-name]\n"
+    "                [-r ram-size-in-MiB]\n"
+    "                [-x number-of-tests] [-u uid-to-use:gid-to-use] [-g gid-to-use]\n"
+    "                [-q] [-f] [-b] [-D] [-p processes | -y]\n"
     "\nVersion: " BON_VERSION "\n");
   exit(1);
 }
@@ -622,8 +701,12 @@ io_error(CPCCHAR message, bool do_exit)
 {
   char buf[1024];
 
-  sprintf(buf, "Bonnie: drastic I/O error (%s)", message);
-  perror(buf);
+  if(!already_printed_error && !do_exit)
+  {
+    sprintf(buf, "Bonnie: drastic I/O error (%s)", message);
+    perror(buf);
+    already_printed_error = 1;
+  }
   if(do_exit)
     exit(1);
   return(1);
